@@ -19,9 +19,11 @@
 import contextlib
 import functools
 
+from orbax.checkpoint import arrays
 from orbax.checkpoint import aggregate_handlers
 from orbax.checkpoint import args
 from orbax.checkpoint import checkpoint_utils
+from orbax.checkpoint import checkpointers
 from orbax.checkpoint import handlers
 from orbax.checkpoint import logging
 from orbax.checkpoint import metadata
@@ -29,10 +31,9 @@ from orbax.checkpoint import msgpack_utils
 from orbax.checkpoint import options
 from orbax.checkpoint import path
 from orbax.checkpoint import serialization
-from orbax.checkpoint import test_utils
 from orbax.checkpoint import transform_utils
 from orbax.checkpoint import tree
-from orbax.checkpoint._src.serialization import type_handlers
+from orbax.checkpoint import type_handlers
 from orbax.checkpoint import utils
 from orbax.checkpoint import version
 # TODO(cpgaffney): Import the public multihost API.
@@ -47,11 +48,7 @@ from orbax.checkpoint.transform_utils import RestoreTransform
 from orbax.checkpoint.transform_utils import Transform
 
 from orbax.checkpoint.abstract_checkpoint_manager import AbstractCheckpointManager
-from orbax.checkpoint.abstract_checkpointer import AbstractCheckpointer
-from orbax.checkpoint.async_checkpointer import AsyncCheckpointer
-from orbax.checkpoint.checkpointer import Checkpointer
-from orbax.checkpoint.pytree_checkpointer import PyTreeCheckpointer
-from orbax.checkpoint.standard_checkpointer import StandardCheckpointer
+from orbax.checkpoint.checkpointers import *
 from orbax.checkpoint.checkpoint_manager import CheckpointManager
 from orbax.checkpoint.checkpoint_manager import AsyncOptions
 from orbax.checkpoint.checkpoint_manager import CheckpointManagerOptions
@@ -80,6 +77,9 @@ from orbax.checkpoint._src.handlers.handler_registration import DefaultCheckpoin
 # warning.
 from orbax.checkpoint._src.handlers.base_pytree_checkpoint_handler import BasePyTreeCheckpointHandler
 
+# Test utils.
+from orbax.checkpoint import test_utils
+
 # A new PyPI release will be pushed everytime `__version__` is increased.
 __version__ = version.__version__
 del version
