@@ -41,6 +41,7 @@ from orbax.checkpoint import utils
 from orbax.checkpoint._src import asyncio_utils
 from orbax.checkpoint._src.handlers import async_checkpoint_handler
 from orbax.checkpoint._src.handlers import base_pytree_checkpoint_handler
+from orbax.checkpoint._src.metadata import empty_values
 from orbax.checkpoint._src.metadata import tree as tree_metadata
 from orbax.checkpoint._src.serialization import serialization
 from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
@@ -184,6 +185,7 @@ def _get_restore_parameters(
     param_names: Optional[PyTree],
     transforms: Optional[PyTree],
     restore_args: Optional[PyTree],
+    pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
     byte_limiter: Optional[LimitInFlightBytes] = None,
     transforms_default_to_original: bool = True,
     use_zarr3: bool = False,
@@ -221,6 +223,7 @@ def _get_restore_parameters(
     restore_args: User-provided restoration arguments. If None, they were not
       provided. Otherwise, the tree has the same structure as the desired output
       tree.
+    pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
     byte_limiter: A LimitInFlightBytes object.
     transforms_default_to_original: See transform_utils.apply_transformations.
     use_zarr3: If True, use Zarr ver3 otherwise Zarr ver2
@@ -246,7 +249,9 @@ def _get_restore_parameters(
       name: str,
       meta_or_value: Union[Any, tree_metadata.ValueMetadataEntry],
   ) -> Union[ParamInfo, Any]:
-    if type_handlers.is_supported_empty_value(meta_or_value):
+    if empty_values.is_supported_empty_value(
+        meta_or_value, pytree_metadata_options
+    ):
       # Empty node, ParamInfo should not be returned.
       return meta_or_value
     elif not isinstance(meta_or_value, tree_metadata.ValueMetadataEntry):
@@ -268,9 +273,8 @@ def _get_restore_parameters(
   if transforms is None:
     for key, meta in flat_structure.items():
       flat_param_infos[key] = _get_param_info(flat_param_names[key], meta)
-    restore_args = tree_utils.serialize_tree(
-        restore_args,
-        keep_empty_nodes=True,
+    restore_args = tree_metadata.serialize_tree(
+        restore_args, pytree_metadata_options
     )
   else:
     if item is None:
@@ -465,6 +469,9 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
       multiprocessing_options: options_lib.MultiprocessingOptions = options_lib.MultiprocessingOptions(),
       type_handler_registry: TypeHandlerRegistry = type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY,
       handler_impl: Optional[BasePyTreeCheckpointHandler] = None,
+      pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
+          tree_metadata.PYTREE_METADATA_OPTIONS
+      ),
   ):
     """Creates PyTreeCheckpointHandler.
 
@@ -485,9 +492,11 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
       type_handler_registry: a type_handlers.TypeHandlerRegistry. If not
         specified, the global type handler registry will be used.
       handler_impl: Allows overriding the internal implementation.
+      pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
     """
     self._aggregate_handler = MsgpackHandler(
-        primary_host=multiprocessing_options.primary_host
+        primary_host=multiprocessing_options.primary_host,
+        pytree_metadata_options=pytree_metadata_options,
     )
     if aggregate_filename is None:
       aggregate_filename = _CHECKPOINT_FILE
@@ -505,7 +514,9 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
         use_zarr3=use_zarr3,
         multiprocessing_options=multiprocessing_options,
         type_handler_registry=type_handler_registry,
+        pytree_metadata_options=pytree_metadata_options,
     )
+    self._pytree_metadata_options = pytree_metadata_options
 
   async def async_save(
       self,
@@ -723,6 +734,11 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
       ValueError: `transforms` is provided without `item`.
       ValueError: `transforms` contains elements with `multi_value_fn`.
     """
+    if self._pytree_metadata_options.support_rich_types:
+      raise NotImplementedError(
+          'Restore is not supported for rich typed metadata yet. Please set'
+          ' PyTreeMetadataOptions.support_rich_types=False.'
+      )
     if not directory.exists():
       raise FileNotFoundError(
           f'Requested directory for restore does not exist at {directory}.'
@@ -784,6 +800,7 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
         self._handler_impl.get_param_names(structure),
         transforms,
         restore_args,
+        self._pytree_metadata_options,
         transforms_default_to_original=transforms_default_to_original,
         use_zarr3=use_zarr3_metadata
         if use_zarr3_metadata is not None
@@ -804,7 +821,7 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
       if (
           isinstance(value_meta, tree_metadata.ValueMetadataEntry)
           and not value_meta.skip_deserialize
-          and value_meta.value_type == type_handlers.RESTORE_TYPE_UNKNOWN
+          and value_meta.value_type == empty_values.RESTORE_TYPE_UNKNOWN
       ):
         return dataclasses.replace(
             value_meta, value_type=type_handlers.default_restore_type(arg)
@@ -878,16 +895,16 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
     """
     # Try reading metadata file.
     try:
-      metadata = self._handler_impl._read_metadata_file(directory)  # pylint: disable=protected-access
-      use_zarr3 = metadata.use_zarr3
-      metadata_tree = metadata.as_nested_tree(keep_empty_nodes=True)
-      flat_metadata = tree_utils.to_flat_dict(
-          metadata_tree, keep_empty_nodes=True
+      internal_tree_metadata = self._handler_impl._read_metadata_file(directory)  # pylint: disable=protected-access
+      use_zarr3 = internal_tree_metadata.use_zarr3
+      value_metadata_tree = internal_tree_metadata.as_nested_tree()
+      flat_value_metadatas = tree_utils.to_flat_dict(
+          value_metadata_tree, keep_empty_nodes=True
       )
     except FileNotFoundError:
       jax.monitoring.record_event('/jax/orbax/deprecation/missing_metadata')
-      metadata_tree = None
-      flat_metadata = None
+      value_metadata_tree = None
+      flat_value_metadatas = None
       use_zarr3 = None
     # Try reading aggregate file.
     try:
@@ -901,32 +918,34 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
       flat_aggregate = None
 
     def _is_empty_value(value):
-      return type_handlers.is_supported_empty_value(
-          value
+      return empty_values.is_supported_empty_value(
+          value, self._pytree_metadata_options
       ) or not utils.leaf_is_placeholder(value)
 
     def _process_aggregate_leaf(value):
       if _is_empty_value(value):
         return value
       return tree_metadata.ValueMetadataEntry(
-          value_type=type_handlers.RESTORE_TYPE_UNKNOWN,
+          value_type=empty_values.RESTORE_TYPE_UNKNOWN,
           skip_deserialize=False,
       )
 
     def _process_metadata_and_aggregate_leaves(value_meta, value):
       if _is_empty_value(value):
         return value
-      if type_handlers.is_empty_typestr(value_meta.value_type):
-        return type_handlers.get_empty_value_from_typestr(value_meta.value_type)
+      if empty_values.is_empty_typestr(value_meta.value_type):
+        return empty_values.get_empty_value_from_typestr(
+            value_meta.value_type, self._pytree_metadata_options
+        )
       return value_meta
 
     # Handle cases of missing metadata and/or aggregate files.
-    structure_tree = metadata_tree or aggregate_tree
-    if flat_metadata is None and flat_aggregate is None:
+    structure_tree = value_metadata_tree or aggregate_tree
+    if flat_value_metadatas is None and flat_aggregate is None:
       raise FileNotFoundError(
           f'No structure could be identified for the checkpoint at {directory}.'
       )
-    elif flat_metadata is None:
+    elif flat_value_metadatas is None:
       # Metadata file is missing. This is an older checkpoint.
       # TODO(b/353310784) Track usages.
       flat_structure = jax.tree_util.tree_map(
@@ -937,22 +956,22 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
     elif flat_aggregate is None:
       # Aggregate file is missing, so we can just use the metadata_tree as the
       # structure. This is a newer checkpoint.
-      return metadata_tree, use_zarr3
+      return value_metadata_tree, use_zarr3
     else:
       # Avoid tree_map because input trees may be mismatched (due to empty
       # values missing from msgpack structure).
       flat_structure = {}
-      for tuple_key in flat_metadata.keys():
-        value_meta = flat_metadata[tuple_key]
+      for tuple_key in flat_value_metadatas.keys():
+        value_meta = flat_value_metadatas[tuple_key]
         if tuple_key in flat_aggregate:
           flat_structure[tuple_key] = _process_metadata_and_aggregate_leaves(
               value_meta, flat_aggregate[tuple_key]
           )
         else:
-          if type_handlers.is_empty_typestr(value_meta.value_type):
+          if empty_values.is_empty_typestr(value_meta.value_type):
             flat_structure[tuple_key] = (
-                type_handlers.get_empty_value_from_typestr(
-                    value_meta.value_type
+                empty_values.get_empty_value_from_typestr(
+                    value_meta.value_type, self._pytree_metadata_options
                 )
             )
           else:
