@@ -12,29 +12,27 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""Array serialization and deserialization.
-
-TODO(b/348434669): De-fork when possible.
-"""
+"""Array serialization and deserialization."""
 
 import asyncio
-from collections.abc import Awaitable
+import collections
+from collections.abc import Mapping
 import contextlib
-import dataclasses
-import functools
 import os
 import re
-from typing import Any, AsyncIterator, Callable, Dict, Optional, Protocol, Sequence, Union
+from typing import Any, AsyncIterator, Dict, Optional, Protocol, Sequence, Union
 
 from absl import logging
 import humanize
 import jax
+from jax.experimental import layout
 import jax.numpy as jnp
 import numpy as np
 from orbax.checkpoint._src.arrays import fragments
-from orbax.checkpoint._src.arrays import numpy_utils
+from orbax.checkpoint._src.arrays import numpy_utils as np_utils
 from orbax.checkpoint._src.arrays import types
 from orbax.checkpoint._src.multihost import multihost
+from orbax.checkpoint._src.serialization import replica_slices
 from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
 import tensorstore as ts
 
@@ -49,27 +47,12 @@ _REMOTE_DRIVER_VALIDATIONS = [
 ]
 
 
-Shape = types.Shape
 Index = types.Index
+Layout = layout.Layout
+Shape = types.Shape
 
 
-async def create_async_array_from_callback(
-    global_shape: Shape,
-    inp_sharding: jax.sharding.Sharding,
-    data_callback: Callable[[Index, jax.Device], Awaitable[jax.Array]],
-) -> jax.Array:
-  device_to_index_map = inp_sharding.devices_indices_map(global_shape)
-  addressable_da = inp_sharding._addressable_device_assignment  # pylint: disable=protected-access
-  future_arrays = [data_callback(device_to_index_map[d], d)
-                   for d in addressable_da]
-  dbs = await asyncio.gather(*future_arrays)
-  return jax.make_array_from_single_device_arrays(
-      global_shape, inp_sharding, dbs
-  )
-
-
-def _get_metadata(arr):
-  local_shape = arr.addressable_data(0).shape
+def _get_metadata(arr: jax.Array, local_shape: Shape):
   return {
       'compressor': {'id': 'zstd'},
       'shape': arr.shape,
@@ -246,136 +229,20 @@ async def reserved_bytes(
     await byte_limiter.release_bytes(nbytes)
 
 
-@dataclasses.dataclass
-class _Shard:
-  index: Index
-  data: np.ndarray | jax.Array
-
-
-def get_unique_shards(
-    arr: jax.Array,
-    replica_id: int | None,
-) -> list[jax.Shard]:
-  """Returns the list of jax.Shard that should be transferred to host memory."""
-  shards_info = ', '.join([
-      f'Shard(index={shard.index}, replica_id={shard.replica_id})'
-      for shard in arr.addressable_shards
-  ])
-  logging.vlog(
-      1,
-      '[process=%d] array shards: replica_id=%d, %s',
-      multihost.process_index(),
-      replica_id,
-      shards_info,
-  )
-  if replica_id is None:
-    replica_id = arr.addressable_shards[0].replica_id
-  return [
-      shard
-      for shard in arr.addressable_shards
-      if shard.replica_id == replica_id
-  ]
-
-
-def _transfer_shard_to_host(
-    shard: jax.Shard,
-    enable_pinned_host_transfer: bool,
-) -> _Shard:
-  """Asynchronously transfers shard data to host memory. Does not block."""
-  data = shard.data
-  has_pinned_host = any(
-      m.kind == 'pinned_host' for m in shard.device.addressable_memories()
-  )
-  if (
-      enable_pinned_host_transfer
-      and has_pinned_host
-      and jax._src.config.enable_memories.value  # pylint: disable=protected-access
-  ):
-    # If available, transfer to pinned host memory
-    sharding = jax.sharding.SingleDeviceSharding(
-        shard.device, memory_kind='pinned_host'
-    )
-    data = jax.device_put(data, sharding)
-  else:
-    data.copy_to_host_async()
-  return _Shard(shard.index, data)
-
-
-def shards_to_fragments(
-    shards: list[jax.Shard | _Shard],
-    global_shape: Shape,
-    dtype: jnp.dtype,
-) -> fragments.Fragments:
-  """Converts a list of jax.Shard to a Fragments."""
-  fragmentses = fragments.Fragments(
-      shape=global_shape,
-      dtype=dtype,
-      fragments=[
-          fragments.Fragment(
-              index=numpy_utils.resolve_slice(shard.index, global_shape),
-              # Conversion to numpy array forces block_until_ready.
-              value=np.asarray(shard.data),
-          )
-          for shard in shards
-      ],
-  )
-  if fragmentses.fragments:
-    fragments.validate_fragments_can_be_stacked(fragmentses)
-  return fragmentses
-
-
-def transfer_arrays_to_host(
-    arrays: Sequence[jax.Array],
-    replica_id: int | None,
-    *,
-    enable_pinned_host_transfer: bool = True,
-) -> Sequence[fragments.Fragments]:
-  """Transfers a jax.Array to host memory. Blocks until completion."""
-  transferred_shards_per_array = []
-  for arr in arrays:
-    shards_to_transfer = get_unique_shards(arr, replica_id)
-    transferred_shards = [
-        _transfer_shard_to_host(shard, enable_pinned_host_transfer)
-        for shard in shards_to_transfer
-    ]
-    transferred_shards_per_array.append(transferred_shards)
-  return [
-      shards_to_fragments(transferred_shards, arr.shape, arr.dtype)
-      for arr, transferred_shards in zip(arrays, transferred_shards_per_array)
-  ]
-
-
-async def _write_fragment(
-    fragment: fragments.Fragment,
-    t: ts.TensorStore,
-    byte_limiter: ByteLimiter,
-):
-  """Writes a single array using TensorStore. No copy is performed."""
-  assert isinstance(fragment.value, np.ndarray)
-  requested_bytes = estimate_write_memory_footprint(fragment.value)
-  async with reserved_bytes(byte_limiter, requested_bytes):
-    await t[fragment.index].write(
-        fragment.value,
-        # Avoid additional copy of input array into the TensorStore chunk
-        # cache. The data array of a shard is guaranteed to be immutable and
-        # therefore it is safe to retain a reference indefinitely.
-        can_reference_source_data_indefinitely=True,
-    )
-
-
 async def async_serialize(
     arr_inp: jax.Array,
     tensorstore_spec: Dict[str, Any],
     context: Optional[ts.Context] = None,
     primary_host: Optional[int] = 0,
     replica_id: int = 0,
+    use_replica_parallel: bool = True,
     transaction: Optional[ts.Transaction] = None,
     byte_limiter: Optional[ByteLimiter] = None,
 ):
   """Serialize an array using TensorStore.
 
   Performs a D2H transfer of the array. Prefer to use
-  `async_serialize_fragments`
+  `async_serialize_from_host`
   by separately performing a D2H transfer, and then starting the serialization
   in a background thread.
 
@@ -388,24 +255,32 @@ async def async_serialize(
       unless you are sure you know what you are doing.
     replica_id: Allows overriding the shard replica id that will be saved. DO
       NOT USE unless you are sure you know what you are doing.
+    use_replica_parallel: Whether to use replica-parallel serialization to allow
+      arrays with replicated shards to be written cooperatively by different
+      hosts.
     transaction: TensorStore transaction to use for opening and writing the
       array.  If not specified, a non-transactional write will be used.
     byte_limiter: A ByteLimiter instance that will be used to limit the number
       of bytes in flight when writing to TensorStore. If None, no limitation
       will be applied.
   """
+  # Start D2H transfer in parallel for each array.
+  rslices = replica_slices.transfer_arrays_to_host(
+      [arr_inp],
+      replica_id,
+      use_replica_parallel,
+  )[0]
+
   byte_limiter = byte_limiter or get_byte_limiter()
   # 'metadata' may not be present at the top level (for example, if we are using
   # a 'cast' driver).
   if not _spec_has_metadata(tensorstore_spec):
-    tensorstore_spec['metadata'] = _get_metadata(arr_inp)
+    tensorstore_spec['metadata'] = _get_metadata(arr_inp, rslices.local_shape)
   # Set dtype if it's not in spec
   if 'dtype' not in tensorstore_spec:
     tensorstore_spec['dtype'] = jnp.dtype(arr_inp.dtype).name
-  # Start D2H transfer in parallel for each array.
-  host_fragments = transfer_arrays_to_host([arr_inp], replica_id)[0]
-  await async_serialize_fragments(
-      host_fragments,
+  await async_serialize_from_host(
+      rslices,
       tensorstore_spec,
       context=context,
       primary_host=primary_host,
@@ -414,8 +289,8 @@ async def async_serialize(
   )
 
 
-async def async_serialize_fragments(
-    shards: fragments.Fragments,
+async def async_serialize_from_host(
+    rslices_on_host: replica_slices.ReplicaSlices,
     tensorstore_spec: Dict[str, Any],
     *,
     context: Optional[ts.Context] = None,
@@ -423,11 +298,10 @@ async def async_serialize_fragments(
     transaction: Optional[ts.Transaction] = None,
     byte_limiter: Optional[ByteLimiter] = None,
 ):
-  """Serialize a host-local shards using TensorStore.
+  """Serialize replica slices using TensorStore.
 
   Args:
-    shards: A fragments.Fragments object. Individual shards are expected to
-      be host-local.
+    rslices_on_host: Replica slices obtained via transfer_arrays_to_host.
     tensorstore_spec: The tensorstore spec to use.
     context: ts.Context instance.
     primary_host: Primary host, which indicates the host that will be treated as
@@ -442,6 +316,8 @@ async def async_serialize_fragments(
   Raises:
     KeyError: If `metadata` or `dtype` is not found in the tensorstore spec.
   """
+  if not rslices_on_host.is_on_host:
+    raise ValueError('Replica slices have not been transferred to host.')
   byte_limiter = byte_limiter or get_byte_limiter()
   if not _spec_has_metadata(tensorstore_spec):
     raise KeyError('`metadata` not found in tensorstore spec.')
@@ -474,15 +350,25 @@ async def async_serialize_fragments(
       context=context,
       transaction=transaction,
   )
-  write_shard_coros = jax.tree.map(
-      functools.partial(
-          _write_fragment,
-          t=t,
-          byte_limiter=byte_limiter,
-      ),
-      shards.fragments,
-  )
-  await asyncio.gather(*write_shard_coros)
+
+  async def write_fragment(fragment: fragments.Fragment):
+    """Writes a single fragment using TensorStore. No copy is performed."""
+    assert isinstance(fragment.value, np.ndarray)
+    requested_bytes = estimate_write_memory_footprint(fragment.value)
+    async with reserved_bytes(byte_limiter, requested_bytes):
+      await t[fragment.index].write(
+          fragment.value,
+          # Avoid additional copy of input array into the TensorStore chunk
+          # cache. The data array of a shard is guaranteed to be immutable and
+          # therefore it is safe to retain a reference indefinitely.
+          can_reference_source_data_indefinitely=True,
+      )
+
+  write_coros = [
+      write_fragment(fragment)
+      for fragment in rslices_on_host.to_fragments().fragments
+  ]
+  await asyncio.gather(*write_coros)
 
 
 def estimate_write_memory_footprint(arr: np.ndarray) -> int:
@@ -523,15 +409,15 @@ def estimate_read_memory_footprint(t: ts.TensorStore,
   return num_bytes
 
 
-async def _read_and_device_put_shard(
-    device: jax.Device,
+async def _read_shard(
     t: ts.TensorStore,
+    *,
     new_shard_shape: Sequence[int],
     dtype: jnp.dtype,
     requested_domain: ts.IndexDomain,
     restricted_domain: ts.IndexDomain,
-) -> jax.Array:
-  """Reads a single shard from TensorStore and places it on device."""
+) -> np.ndarray:
+  """Reads a single shard from TensorStore into host memory."""
   # This maybe needed because the shape the array was saved with is smaller
   # than the requested shape of the array in which it will be reloaded. So
   # the extra values will be filled with 0s.
@@ -543,55 +429,128 @@ async def _read_and_device_put_shard(
     # Cast while reloading on process to avoid 2 copies on device if the
     # casting is done on device.
     out = out.astype(dtype)
-  # Convert to jnp array so that layouts are initialized properly for
-  # sub-byte dtypes.
-  # TODO(yashkatariya): This is a band-aid fix. Figure out a better way to
-  # make this work.
-  if out.dtype == jnp.int4:
-    out = jnp.asarray(out)  # type: ignore
-  return jax.device_put(out, jax.sharding.SingleDeviceSharding(device))
+  return out
 
 
-async def _read_array_index_callback(
+async def _read_array_index_and_device_put(
+    devices: list[jax.Device],
     index: Index,
-    device: jax.Device,
     t: ts.TensorStore,
-    shape: Sequence[int],
-    new_shard_shape: Sequence[int],
+    *,
+    global_shape: Shape,
+    new_shard_shape: Shape,
     dtype: jnp.dtype,
     byte_limiter: ByteLimiter,
     strict: bool,
-) -> jax.Array:
-  """Callback that reads an array index and places on device."""
-  if strict and t.shape != shape:
-    raise ValueError(
-        f'Requested shape: {shape} is not compatible with the stored shape:'
-        f' {t.shape}. Truncating/padding is disabled by setting of'
-        ' `strict=True`. When using standard Orbax APIs, this behavior can be'
-        ' modified by specifying `strict=False` in `ArrayRestoreArgs` for any'
-        ' array in which padding/truncation is desired.'
-    )
-  requested_domain = ts.IndexTransform(input_shape=shape)[index].domain
-  restricted_domain = t.domain.intersect(requested_domain)
+    dll: Optional[layout.DeviceLocalLayout],
+) -> list[jax.Array]:
+  """Callback that reads an array index and places on the devices."""
+  for sl in index:
+    if sl.step is not None and sl.step != 1:
+      raise ValueError(
+          f'Non-contiguous domain for index: {index} not supported. Found:'
+          f' {sl.step}'
+      )
+
+  if strict:
+    if t.shape == global_shape:
+      domain = ts.IndexDomain(shape=global_shape)[ts.d[:][index]]
+      requested_domain = domain
+      restricted_domain = domain
+    else:
+      raise ValueError(
+          f'Requested shape: {global_shape} is not compatible with the stored'
+          f' shape: {t.shape}. Truncating/padding is disabled by setting of'
+          ' `strict=True`. When using standard Orbax APIs, this behavior can'
+          ' be modified by specifying `strict=False` in `ArrayRestoreArgs` for'
+          ' any array in which padding/truncation is desired.'
+      )
+  else:
+    requested_domain = ts.IndexTransform(input_shape=global_shape)[index].domain
+    restricted_domain = t.domain.intersect(requested_domain)
+
   requested_bytes = estimate_read_memory_footprint(t, restricted_domain)
+  result = []
   # Limit the bytes read for every shard.
+  # Perform read for index once, and place it on all relevant devices within
+  # the `reserved_bytes` context.
+  # TODO(b/381111280) This de-duplication of reads does not fully solve the
+  # problem of read amplification, since we can still run into problems if
+  # we are resharding. See b/381111280 for details.
   async with reserved_bytes(byte_limiter, requested_bytes):
-    result = await _read_and_device_put_shard(
-        device,
-        t,
-        new_shard_shape,
-        dtype,
-        requested_domain,
-        restricted_domain,
-    )
+    try:
+      shard = await _read_shard(
+          t=t,
+          new_shard_shape=new_shard_shape,
+          dtype=dtype,
+          requested_domain=requested_domain,
+          restricted_domain=restricted_domain,
+      )
+    except BaseException as e:
+      raise Exception(  # pylint: disable=broad-exception-raised
+          f'Encountered error while reading array index: {index}. See full'
+          f' TensorStore details: {t.spec}.'
+      ) from e
+    for device in devices:
+      result.append(
+          jax.device_put(
+              shard, Layout(dll, jax.sharding.SingleDeviceSharding(device))
+          )
+      )
   return result
 
 
+def _get_device_to_index_map(
+    global_shape: Shape, sharding: jax.sharding.Sharding
+) -> Mapping[jax.Device, Index]:
+  return sharding.devices_indices_map(global_shape)
+
+
+async def read_and_create_array(
+    t: ts.TensorStore,
+    *,
+    global_shape: Shape,
+    new_shard_shape: Shape,
+    sharding: jax.sharding.Sharding,
+    dtype: jnp.dtype,
+    byte_limiter: ByteLimiter,
+    strict: bool,
+    dll: Optional[layout.DeviceLocalLayout],
+) -> jax.Array:
+  """Read shards from TensorStore and create a jax.Array."""
+  local_indices_devices_map: dict[types.HashableIndex, list[jax.Device]] = (
+      collections.defaultdict(list)
+  )
+  for d, idx in _get_device_to_index_map(global_shape, sharding).items():
+    if d in sharding._addressable_device_assignment:  # pylint: disable=protected-access
+      local_indices_devices_map[
+          np_utils.to_hashable_index(idx, shape=global_shape)
+      ].append(d)
+
+  read_array_coros = [
+      _read_array_index_and_device_put(
+          devices,
+          np_utils.from_hashable_index(idx),
+          t,
+          global_shape=global_shape,
+          new_shard_shape=new_shard_shape,
+          dtype=dtype,
+          byte_limiter=byte_limiter,
+          strict=strict,
+          dll=dll,
+      )
+      for idx, devices in local_indices_devices_map.items()
+  ]
+  dbs = sum(await asyncio.gather(*read_array_coros), [])
+  return jax.make_array_from_single_device_arrays(global_shape, sharding, dbs)
+
+
 async def async_deserialize(
-    user_in_sharding: jax.sharding.Sharding,
+    user_sharding: jax.sharding.Sharding | Layout,
     tensorstore_spec: Union[ts.Spec, Dict[str, Any]],
-    global_shape: Optional[Sequence[int]] = None,
+    global_shape: Optional[Shape] = None,
     dtype: Optional[jnp.dtype] = None,
+    *,
     byte_limiter: Optional[ByteLimiter] = None,
     context: Optional[ts.Context] = None,
     assume_metadata: bool = False,
@@ -600,29 +559,32 @@ async def async_deserialize(
   """Reads an array using TensorStore."""
   byte_limiter = byte_limiter or get_byte_limiter()
   context = context or ts_utils.get_ts_context(use_ocdbt=False)
-  in_sharding = user_in_sharding
-  if not isinstance(in_sharding, jax.sharding.Sharding):
+  sharding = (
+      user_sharding.sharding
+      if isinstance(user_sharding, Layout)
+      else user_sharding
+  )
+  if not isinstance(sharding, jax.sharding.Sharding):
     raise ValueError(
         'sharding passed to deserialization should be specified, concrete and'
-        f' an instance of `jax.sharding.Sharding`. Got {in_sharding}')
+        f' an instance of `jax.sharding.Sharding`. Got {sharding}'
+    )
+  dll = sharding.device_local_layout if isinstance(sharding, Layout) else None
   t = await ts.open(
       tensorstore_spec,
       open=True,
       assume_metadata=assume_metadata,
       context=context,
   )
-  shape = t.shape if global_shape is None else global_shape
-  new_shard_shape = in_sharding.shard_shape(tuple(shape))
-  return await create_async_array_from_callback(
-      tuple(shape),
-      in_sharding,
-      functools.partial(
-          _read_array_index_callback,
-          t=t,
-          shape=shape,
-          new_shard_shape=new_shard_shape,
-          dtype=dtype,
-          byte_limiter=byte_limiter,
-          strict=strict,
-      ),
+  global_shape = tuple(t.shape if global_shape is None else global_shape)
+  new_shard_shape = sharding.shard_shape(global_shape)
+  return await read_and_create_array(
+      t,
+      global_shape=global_shape,
+      new_shard_shape=new_shard_shape,
+      sharding=sharding,
+      dtype=dtype,
+      byte_limiter=byte_limiter,
+      strict=strict,
+      dll=dll,
   )
