@@ -30,21 +30,22 @@ from etils import epy
 import jax
 from jax.experimental.array_serialization import serialization as jax_serialization
 from orbax.checkpoint import abstract_checkpoint_manager
-from orbax.checkpoint import abstract_checkpointer
 from orbax.checkpoint import args as args_lib
-from orbax.checkpoint import async_checkpointer
 from orbax.checkpoint import checkpoint_args
-from orbax.checkpoint import checkpointer as checkpointer_lib
 from orbax.checkpoint import options as options_lib
 from orbax.checkpoint import utils
+from orbax.checkpoint._src.checkpointers import abstract_checkpointer
+from orbax.checkpoint._src.checkpointers import async_checkpointer
+from orbax.checkpoint._src.checkpointers import checkpointer as checkpointer_lib
 from orbax.checkpoint._src.handlers import checkpoint_handler
 from orbax.checkpoint._src.handlers import composite_checkpoint_handler
 from orbax.checkpoint._src.handlers import handler_registration
 from orbax.checkpoint._src.handlers import json_checkpoint_handler
 from orbax.checkpoint._src.handlers import proto_checkpoint_handler
 from orbax.checkpoint._src.metadata import checkpoint
+from orbax.checkpoint._src.metadata import root_metadata_serialization
 from orbax.checkpoint._src.multihost import multihost
-from orbax.checkpoint._src.path import atomicity
+from orbax.checkpoint._src.path import atomicity_types
 from orbax.checkpoint._src.path import deleter
 from orbax.checkpoint._src.path import step as step_lib
 from orbax.checkpoint._src.path import utils as path_utils
@@ -74,6 +75,8 @@ CheckpointHandler = checkpoint_handler.CheckpointHandler
 CheckpointArgs = checkpoint_args.CheckpointArgs
 CheckpointHandlersDict = Mapping[str, CheckpointHandler]
 CheckpointHandlerRegistry = handler_registration.CheckpointHandlerRegistry
+serialize_root_metadata = root_metadata_serialization.serialize
+deserialize_root_metadata = root_metadata_serialization.deserialize
 
 AsyncOptions = options_lib.AsyncOptions
 MultiprocessingOptions = options_lib.MultiprocessingOptions
@@ -190,6 +193,12 @@ class CheckpointManagerOptions:
   keep_period:
     If set, any existing checkpoints matching checkpoint_step % keep_period == 0
     will not be deleted.
+  should_keep_fn:
+    WARNING: It is an experimental feature and may change without notice.
+    If set then this Callable overrides the behavior of `keep_period`, which is
+    ignored. It does not change the behavior of `keep_time_interval` and
+    `keep_checkpoints_without_metrics`. It is a predicate with signature:
+    `Callable[[int], bool]`, where int is the step number of the checkpoint.
   best_fn:
     If set, maintains checkpoints based on the quality of given
     metrics rather than recency. The function should accept a PyTree of metrics,
@@ -250,12 +259,18 @@ class CheckpointManagerOptions:
     ignored.
   file_options: Options to configure checkpoint directories and files.
     default=FileOptions().
+  save_root_metadata: If True, saves root-level metadata about checkpoints.
+    This metadata is not step-specific and is written only once.
+  temporary_path_class:
+    Optional. The concrete `atomicity_types.TemporaryPath` class to be used by
+    the underlying `Checkpointer`.
   """
 
   save_interval_steps: int = 1
   max_to_keep: Optional[int] = None
   keep_time_interval: Optional[datetime.timedelta] = None
   keep_period: Optional[int] = None
+  should_keep_fn: Optional[Callable[[int], bool]] = None
   best_fn: Optional[Callable[[PyTree], float]] = None
   best_mode: str = 'max'
   keep_checkpoints_without_metrics: bool = True
@@ -276,7 +291,8 @@ class CheckpointManagerOptions:
   )
   should_save_fn: Optional[Callable[[int, Optional[int]], bool]] = None
   file_options: FileOptions = dataclasses.field(default_factory=FileOptions)
-  temporary_path_class: Optional[Type[atomicity.TemporaryPath]] = None
+  save_root_metadata: bool = True
+  temporary_path_class: Optional[Type[atomicity_types.TemporaryPath]] = None
 
   def __post_init__(self):
     if self.best_mode not in ('min', 'max'):
@@ -306,8 +322,10 @@ class CheckpointManagerOptions:
       )
     if self.read_only and self.keep_period is not None:
       self.keep_period = None
+      self.should_keep_fn = None
       logging.warning(
-          'CheckpointManagerOptions.read_only=True, setting keep_period=None.'
+          'CheckpointManagerOptions.read_only=True, setting keep_period=None'
+          ' and should_keep_fn=None.'
       )
     if self.read_only and self.create:
       self.create = False
@@ -337,6 +355,13 @@ class CheckpointManagerOptions:
           'CheckpointManagerOptions.read_only=True, setting'
           ' should_save_fn=None.'
       )
+    if self.should_keep_fn is not None:
+      logging.warning(
+          'CheckpointManagerOptions.should_keep_fn is set, setting'
+          ' keep_period=None (was %s).',
+          self.keep_period,
+      )
+      self.keep_period = None
     self.save_on_steps = frozenset(self.save_on_steps or ())
 
 
@@ -407,14 +432,14 @@ def _create_root_directory(
       multihost.unique_barrier_key(
           'CheckpointManager:create_directory',
           prefix=multiprocessing_options.barrier_sync_key_prefix,
-          suffix=None,
+          # suffix=None,
       ),
       timeout=multihost.DIRECTORY_CREATION_TIMEOUT,
       processes=multiprocessing_options.active_processes,
   )
 
 
-def _determine_default_item_mode_from_args(
+def determine_default_item_mode_from_args(
     args: args_lib.CheckpointArgs,
 ) -> bool:
   if isinstance(args, args_lib.Composite):
@@ -593,14 +618,12 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       )
 
     # For async_checkpointer.
-    self._non_blocking_checkpoint_metadata_store = (
-        checkpoint.checkpoint_metadata_store(enable_write=True)
+    self._non_blocking_metadata_store = (
+        checkpoint.metadata_store(enable_write=True)
     )
     # For metadata checkpointer and regular checkpointer.
-    self._blocking_checkpoint_metadata_store = (
-        checkpoint.checkpoint_metadata_store(
-            enable_write=True, blocking_write=True
-        )
+    self._blocking_metadata_store = (
+        checkpoint.metadata_store(enable_write=True, blocking_write=True)
     )
 
     if checkpointers is not None:
@@ -684,21 +707,13 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
     self._checkpoints = self._load_checkpoint_infos()
 
-    self._metadata_checkpointer = Checkpointer(
-        JsonCheckpointHandler(
-            multiprocessing_options=self._multiprocessing_options
-        ),
-        multiprocessing_options=self._options.multiprocessing_options,
-        file_options=self._options.file_options,
-        checkpoint_metadata_store=self._blocking_checkpoint_metadata_store,
-        temporary_path_class=self._options.temporary_path_class,
-    )
-    if self._options.read_only and not self._metadata_path().exists():
+    self._metadata_dir = self.directory / METADATA_ITEM_NAME
+    if self._options.read_only and not self._metadata_dir.exists():
       self._metadata = {} if metadata is None else metadata
     else:
       self._metadata = None
-    if metadata is not None and not self._options.read_only:
-      self._save_metadata(metadata)
+
+    self._maybe_save_metadata(metadata)
 
     # TODO: b/359854428 - Move Finalize biz logic to a separate class/module.
     self._finalize_thread_lock = threading.Lock()
@@ -760,7 +775,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
           multiprocessing_options=options.multiprocessing_options,
           async_options=options.async_options or AsyncOptions(),
           file_options=options.file_options,
-          checkpoint_metadata_store=self._non_blocking_checkpoint_metadata_store,
+          checkpoint_metadata_store=self._non_blocking_metadata_store,
           temporary_path_class=options.temporary_path_class,
       )
     else:
@@ -768,7 +783,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
           handler,
           multiprocessing_options=options.multiprocessing_options,
           file_options=options.file_options,
-          checkpoint_metadata_store=self._blocking_checkpoint_metadata_store,
+          checkpoint_metadata_store=self._blocking_metadata_store,
           temporary_path_class=options.temporary_path_class,
       )
 
@@ -1046,7 +1061,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         multihost.unique_barrier_key(
             'CheckpointManager:deleted_step',
             prefix=self._multiprocessing_options.barrier_sync_key_prefix,
-            suffix=str(step),
+            # suffix=str(step),
         ),
         timeout=multihost.DIRECTORY_DELETION_TIMEOUT,
         processes=self._multiprocessing_options.active_processes,
@@ -1103,7 +1118,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     if items is None and args is None:
       raise ValueError('Must provide `args` for `save`.')
     if self._default_item is None:
-      self._default_item = _determine_default_item_mode_from_args(args)
+      self._default_item = determine_default_item_mode_from_args(args)
     self._validate_args(items, args)
     if not force and not self.should_save(step):
       return False
@@ -1207,7 +1222,6 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
           multihost.unique_barrier_key(
               'CheckpointManager:delete_unfinalized_step_gcs',
               prefix=self._multiprocessing_options.barrier_sync_key_prefix,
-              suffix=str(step),
           ),
           timeout=multihost.DIRECTORY_DELETION_TIMEOUT,
           processes=self._multiprocessing_options.active_processes,
@@ -1243,7 +1257,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         multihost.unique_barrier_key(
             'CheckpointManager:old_steps_to_remove',
             prefix=self._multiprocessing_options.barrier_sync_key_prefix,
-            suffix=str(step),
+            # suffix=str(step),
         ),
         processes=self._multiprocessing_options.active_processes,
     )
@@ -1308,6 +1322,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     """See superclass documentation."""
     if step is None:
       step = self.latest_step()
+      if step is None:
+        raise FileNotFoundError(f'No steps found in {self.directory}.')
     directory = directory or self.directory
     directory = epath.Path(directory)
     step_stats = step_statistics.RestoreStepStatistics()
@@ -1479,31 +1495,57 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         )
     )
 
-  def _metadata_path(self) -> epath.Path:
-    return self.directory / METADATA_ITEM_NAME
-
-  def _save_metadata(self, metadata: Mapping[str, Any]):
-    """Saves CheckpointManager level metadata, skips if already present."""
-    path = self._metadata_path()
-    path_exists = path.exists()
-    # Ensure that we check across all processes before potentially saving.
-    multihost.sync_global_processes(
-        multihost.unique_barrier_key(
-            'CheckpointManager:check_metadata_existence',
-            prefix=self._multiprocessing_options.barrier_sync_key_prefix,
-            suffix=self.directory.name,
-        ),
-        processes=self._multiprocessing_options.active_processes,
+  def _metadata_file_path(self, legacy: bool = False) -> epath.Path:
+    if not self._metadata_dir.exists():
+      raise ValueError('Metadata directory is not initialized.')
+    return checkpoint.root_metadata_file_path(
+        self._metadata_dir, legacy=legacy
     )
-    if not path_exists:  # May have been created by a previous run.
-      self._metadata_checkpointer.save(path, metadata)
+
+  def _maybe_save_metadata(self, metadata: Mapping[str, Any]):
+    """Saves CheckpointManager level metadata, skips if already present."""
+    if self._options.save_root_metadata:
+      logging.info('Saving root metadata')
+      if (metadata is not None and
+          not self._options.read_only and
+          utils.is_primary_host(self._multiprocessing_options.primary_host)):
+        logging.info('Creating metadata directory')
+        self._metadata_dir.mkdir(parents=True, exist_ok=True)
+        file_path = self._metadata_file_path()
+        if not file_path.exists():  # May have been created by a previous run.
+          logging.info('Writing root metadata')
+          metadata_to_save = checkpoint.RootMetadata(
+              format=None,
+              custom=dict(metadata),
+          )
+          self._blocking_metadata_store.write(
+              file_path, serialize_root_metadata(metadata_to_save)
+          )
+
+      multihost.sync_global_processes(
+          multihost.unique_barrier_key(
+              'CheckpointManager:save_metadata',
+              prefix=self._multiprocessing_options.barrier_sync_key_prefix,
+          ),
+          processes=self._multiprocessing_options.active_processes,
+      )
 
   def metadata(self) -> Mapping[str, Any]:
     """See superclass documentation."""
     if self._metadata is None:
-      path = self._metadata_path()
-      if path.exists():
-        self._metadata = self._metadata_checkpointer.restore(path)
+      if self._metadata_dir.exists():
+        file_path = self._metadata_file_path()
+        if not file_path.exists():
+          logging.warning('New metadata file not found in directory: %s. '
+                          'Will try to read legacy metadata file.',
+                          self._metadata_dir)
+          file_path = self._metadata_file_path(legacy=True)
+        serialized_metadata = self._blocking_metadata_store.read(file_path)
+        self._metadata = deserialize_root_metadata(serialized_metadata).custom
+        if self._metadata is None:
+          raise FileNotFoundError(
+              f'Failed to read metadata from {file_path}.'
+          )
       else:
         self._metadata = {}
     return self._metadata
@@ -1614,6 +1656,15 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
           )
           kept_checkpoints.add(info)
           continue
+      if (
+          self._options.should_keep_fn is not None
+          and self._options.should_keep_fn(info.step)
+      ):
+        logging.info(
+            'Preserving %s: (Reason: on should_keep_fn callback).', info
+        )
+        kept_checkpoints.add(info)
+        continue
       if (
           self._options.keep_period is not None
           and info.step % self._options.keep_period == 0
@@ -1765,7 +1816,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
   def _finalize(self, step: int, steps_to_remove: List[int]):
     """Finalizes individual items and starts garbage collection."""
     process_index = multihost.process_index()
-    self._non_blocking_checkpoint_metadata_store.wait_until_finished()
+    self._non_blocking_metadata_store.wait_until_finished()
     self._wait_for_checkpointers()
     # If an error is encountered while waiting for commit futures to complete,
     # we will not proceed past this point.
@@ -1804,8 +1855,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     self.wait_until_finished()
     self._checkpointer.close()
     # Call after checkpointer.close().
-    self._non_blocking_checkpoint_metadata_store.close()
-    self._blocking_checkpoint_metadata_store.close()
+    self._non_blocking_metadata_store.close()
+    self._blocking_metadata_store.close()
     self._checkpoint_deleter.close()
 
   def __contextmanager__(
