@@ -15,7 +15,7 @@
 """A class providing emergency checkpoint management.
 
 
-This class is experimental; do not use without specific approval.
+WARNING: This class is experimental; do not use without specific approval.
 
 NOTE: All classes within this module should be called across all *relevant*
 processes. CheckpointManager is designed to be created and called across
@@ -25,11 +25,12 @@ intended to work only with the persistent checkpoint on the primary slice should
 always be called across all processes within the primary slice.
 """
 
+import asyncio
 import collections
 import dataclasses
 import enum
 import functools
-import json
+import itertools
 import operator
 import time
 from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Set
@@ -43,14 +44,14 @@ import numpy as np
 from orbax.checkpoint import abstract_checkpoint_manager
 from orbax.checkpoint import args as args_lib
 from orbax.checkpoint import checkpoint_manager
-from orbax.checkpoint import utils
+from orbax.checkpoint import checkpoint_utils
 from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
-from orbax.checkpoint._src.multihost import counters
 from orbax.checkpoint._src.multihost import multihost
 from orbax.checkpoint._src.multihost import multislice
 from orbax.checkpoint._src.path import step as step_lib
 from orbax.checkpoint._src.serialization import type_handlers
-from orbax.checkpoint.experimental.emergency import multihost as emergency_multihost
+from orbax.checkpoint.experimental.emergency import local_checkpoint_data_debugging
+from orbax.checkpoint.experimental.emergency import mesh_consistency
 from orbax.checkpoint.logging import abstract_logger
 from orbax.checkpoint.logging import standard_logger
 from orbax.checkpoint.logging import step_statistics
@@ -62,103 +63,17 @@ CheckpointHandler = checkpoint_manager.CheckpointHandler
 P = jax.sharding.PartitionSpec
 PyTreeCheckpointHandler = pytree_checkpoint_handler.PyTreeCheckpointHandler
 unique_barrier_key = multihost._unique_barrier_key  # pylint: disable=protected-access
-
-_PROCESS_METADATA_FOLDER = 'process_metadata'
-_PROCESS_METADATA_FILE_NAME = 'process_metadata.json'
-_GLOBAL_PROCESS_METADATA_FILE_NAME = 'global_process_metadata.json'
-_MESH_METADATA_FILE_NAME = 'mesh_metadata.json'
+ChunkId = local_checkpoint_data_debugging.ChunkId
+get_present_and_missing_chunks = (
+    local_checkpoint_data_debugging.get_present_and_missing_chunks
+)
 
 _PRIMARY_REPLICA_ID = 0
 _SECONDARY_REPLICA_ID = 1
 
 
-def _write_process_metadata(path: epath.Path, mesh: jax.sharding.Mesh):
-  """Write process metadata to the given path."""
-  logging.info('Saving process index metadata at %s', path)
-
-  if multihost.process_index() == 0:
-    path.mkdir(parents=False, exist_ok=False)
-    runtime_to_distributed_ids = (
-        emergency_multihost.runtime_to_distributed_ids()
-    )
-    (path / _GLOBAL_PROCESS_METADATA_FILE_NAME).write_text(
-        json.dumps(runtime_to_distributed_ids)
-    )
-    (path / _MESH_METADATA_FILE_NAME).write_text(
-        json.dumps([int(id) for id in mesh.device_ids.flatten()])
-    )
-  multihost.sync_global_processes('create_process_metadata')
-
-
-def _read_process_metadata(path: epath.Path):
-  """Read process metadata from the given path."""
-  logging.info('Loading process index metadata from %s', path)
-
-  runtime_to_distributed_ids = json.loads(
-      (path / _GLOBAL_PROCESS_METADATA_FILE_NAME).read_text()
-  )
-  device_ids = json.loads((path / _MESH_METADATA_FILE_NAME).read_text())
-  return runtime_to_distributed_ids, device_ids
-
-
-def _maybe_save_process_metadata(
-    path: epath.Path, global_mesh: jax.sharding.Mesh
-) -> bool:
-  """Saves process metadata if it does not already exist."""
-  metadata_folder = path / _PROCESS_METADATA_FOLDER
-  if metadata_folder.exists():
-    return False
-  # All processes must check the folder before proceeding. Otherwise the
-  # primary process may create the folder from scratch before another
-  # process has a chance to check it.
-  multihost.sync_global_processes('check_process_metadata_folder')
-  _write_process_metadata(metadata_folder, global_mesh)
-  return True
-
-
-def _should_restore_mesh_from_metadata(path: epath.Path) -> bool:
-  metadata_path = path / _PROCESS_METADATA_FOLDER
-  return metadata_path.exists()
-
-
-def _consistent_restore_mesh_from_metadata(
-    path: epath.Path, global_mesh: jax.sharding.Mesh
-) -> jax.sharding.Mesh:
-  """Create a mesh consistent with the saved metadata."""
-  metadata_path = path / _PROCESS_METADATA_FOLDER
-  runtime_to_distributed_ids, device_ids = _read_process_metadata(metadata_path)
-  assert isinstance(device_ids, list)
-  logging.info(
-      'From process metadata, runtime_to_distributed_ids=%s',
-      runtime_to_distributed_ids,
-  )
-  logging.info('From process metadata, device_ids=%s', device_ids)
-  consistent_mesh = emergency_multihost.consistent_restore_mesh(
-      global_mesh, device_ids, runtime_to_distributed_ids
-  )
-  logging.info(
-      'Created consistent mesh with device_ids=%s',
-      consistent_mesh.device_ids.flatten(),
-  )
-  return consistent_mesh
-
-
-def local_checkpoint_handler() -> PyTreeCheckpointHandler:
-  """Create a PyTreeCheckpointHandler for local checkpoints."""
-  local_registry = type_handlers.create_type_handler_registry(
-      (
-          jax.Array,
-          type_handlers.ArrayHandler(primary_host=None, replica_id=None),
-      ),
-  )
-  return PyTreeCheckpointHandler(
-      use_ocdbt=True,
-      use_zarr3=True,
-      multiprocessing_options=checkpoint_manager.MultiprocessingOptions(
-          primary_host=None,
-      ),
-      type_handler_registry=local_registry,
-  )
+local_all_steps_broadcast_counter = itertools.count()
+find_complete_slice_broadcast_counter = itertools.count()
 
 
 def _local_checkpoint_handler(
@@ -173,7 +88,11 @@ def _local_checkpoint_handler(
   local_registry = type_handlers.create_type_handler_registry(
       (
           jax.Array,
-          type_handlers.ArrayHandler(primary_host=None, replica_id=None),
+          type_handlers.ArrayHandler(
+              primary_host=None,
+              replica_id=None,
+              use_replica_parallel=False,
+          ),
       ),
   )
   return PyTreeCheckpointHandler(
@@ -195,7 +114,9 @@ def _persistent_checkpoint_handler(
       (
           jax.Array,
           type_handlers.ArrayHandler(
-              primary_host=multiprocessing_options.primary_host, replica_id=0
+              primary_host=multiprocessing_options.primary_host,
+              replica_id=0,
+              use_replica_parallel=False,
           ),
       ),
   )
@@ -318,17 +239,14 @@ class CheckpointManagerOptions:
 class _BarrierIdentifier(enum.Enum):
   """Identifies the barrier being run."""
 
-  GLOBAL_MAX = 'global_max'
   LOCAL_ALL_STEPS = 'local_all_steps'
   FIND_COMPLETE_SLICE = 'find_complete_slice'
 
   def get_counter(self) -> str:
-    if self.name == self.GLOBAL_MAX.name:
-      return counters.global_max_broadcast_counter()
-    elif self.name == self.LOCAL_ALL_STEPS.name:
-      return counters.local_all_steps_broadcast_counter()
+    if self.name == self.LOCAL_ALL_STEPS.name:
+      return str(next(local_all_steps_broadcast_counter))
     elif self.name == self.FIND_COMPLETE_SLICE.name:
-      return counters.find_complete_slice_broadcast_counter()
+      return str(next(find_complete_slice_broadcast_counter))
     else:
       raise ValueError(f'Unknown barrier identifier: {self.name}')
 
@@ -419,7 +337,7 @@ def _process_local_to_global(
       f'{barrier_name}_{barrier_id.get_counter()}_{barrier_processes_as_string}'
   )
 
-  client = multihost._get_jax_distributed_client()  # pylint: disable=protected-access
+  client = multihost.get_jax_distributed_client()
   broadcast_dir_key = f'broadcast_{barrier_name_and_id}/'
   broadcast_dir_key = unique_barrier_key(broadcast_dir_key) + '/'
   broadcast_key = broadcast_dir_key + str(multihost.process_index())
@@ -526,6 +444,7 @@ class _LocalCheckpointManager(checkpoint_manager.CheckpointManager):
         single_host_load_and_broadcast=False,
         # enable_background_delete set to False to ensure gc is done before save
         enable_background_delete=False,
+        save_root_metadata=False,
     )
     self._logger = logger or standard_logger.StandardLogger()
     self._coordination_timeout_secs = (
@@ -538,18 +457,31 @@ class _LocalCheckpointManager(checkpoint_manager.CheckpointManager):
         item_handlers=_local_checkpoint_handler(multiprocessing_options),
         logger=self._logger,
     )
+    self._run_initial_garbage_collection()
+    # Set additional properties.
+    self._max_to_keep = options.local.max_to_keep
+    self._local_options = options.local
+    self._steps = list(self.all_steps(read=True))
 
-    # Remove steps that might be left over from previous runs.
+  def _run_initial_garbage_collection(self):
+    """Remove steps that might be left over from previous runs."""
     steps_to_remove = self._get_old_steps_to_remove()
     self._checkpoints = [
         info for info in self._checkpoints if info.step not in steps_to_remove
     ]
     self._checkpoint_deleter.delete_steps(steps_to_remove)
 
-    # Set additional properties.
-    self._max_to_keep = options.local.max_to_keep
-    self._local_options = options.local
-    self._steps = list(self.all_steps(read=True))
+    checkpoint_steps_set = set([info.step for info in self._checkpoints])
+    if mesh_consistency.process_metadata_folder(self.directory).exists():
+      metadata_paths_to_remove = [
+          p
+          for p in mesh_consistency.process_metadata_folder(
+              self.directory
+          ).iterdir()
+          if int(p.name) not in checkpoint_steps_set
+      ]
+      for p in metadata_paths_to_remove:
+        p.rmtree()
 
   def local_host_steps(self, read: bool) -> Sequence[int]:
     """Returns steps known to local host."""
@@ -632,6 +564,14 @@ class _LocalCheckpointManager(checkpoint_manager.CheckpointManager):
     if saved:
       self._steps.append(step)
       self._steps = self._steps[-self._max_to_keep :]
+      # TODO(cpgaffney) make this asynchronous. Not a priority because the write
+      # happens to local storage, so it should be fast.
+      mesh_consistency.save_process_metadata(
+          self.directory,
+          step,
+          self._global_mesh,
+          multihost.distributed_to_device_ids(),
+      )
     return saved
 
   def reload(self):
@@ -643,6 +583,28 @@ class _LocalCheckpointManager(checkpoint_manager.CheckpointManager):
     self._steps = list(self.all_steps(read=True))
 
 
+def _get_single_slice_sharding(
+    mesh: jax.sharding.Mesh,
+    pspec: jax.sharding.PartitionSpec,
+    replica_id: int,
+    replica_axis_index: int,
+):
+  """Get sharding for a single slice."""
+  slice_devices = multislice.slice_devices(
+      mesh,
+      replica_id=replica_id,
+      replica_axis_index=replica_axis_index,
+  )
+  single_slice_mesh_shape = [
+      1 if i == replica_axis_index else d
+      for i, d in enumerate(mesh.devices.shape)
+  ]
+  slice_mesh = jax.sharding.Mesh(
+      slice_devices.reshape(single_slice_mesh_shape), mesh.axis_names
+  )
+  return jax.sharding.NamedSharding(slice_mesh, pspec)
+
+
 class CheckpointManager(
     abstract_checkpoint_manager.AbstractCheckpointManager, epy.ContextManager
 ):
@@ -692,6 +654,17 @@ class CheckpointManager(
   ):
     self._local_directory = epath.Path(local_directory)
     self._persistent_directory = epath.Path(persistent_directory)
+    if not self._local_directory.exists():
+      raise FileNotFoundError(
+          f'Local directory {self._local_directory} must be created by the'
+          ' caller.'
+      )
+    if not self._persistent_directory.exists():
+      raise FileNotFoundError(
+          f'Persistent directory {self._persistent_directory} must be created'
+          ' by the caller.'
+      )
+
     self._logger = logger or standard_logger.StandardLogger()
     # TODO: b/330585086 - Fully support options.
     options = options or CheckpointManagerOptions()
@@ -703,7 +676,6 @@ class CheckpointManager(
         self._replica_axis_index,
         self._global_mesh.axis_names[self._replica_axis_index],
     )
-    _maybe_save_process_metadata(self._persistent_directory, self._global_mesh)
 
     self._abstract_state = abstract_state
     self._slice_id = multislice.process_slice_id(
@@ -807,6 +779,7 @@ class CheckpointManager(
         multihost.process_index(),
         jax.process_index(),
     )
+    logging.vlog(1, 'Local devices: %s', jax.local_devices())
 
   def _cleanup_local_tmp_directories(self):
     logging.info(
@@ -832,6 +805,7 @@ class CheckpointManager(
         multiprocessing_options=persistent_multiprocessing_options,
         enable_async_checkpointing=self._options.enable_async_checkpointing,
         should_save_fn=self._options.persistent.should_save_fn,
+        save_root_metadata=False,
     )
     return checkpoint_manager.CheckpointManager(
         self._persistent_directory,
@@ -921,7 +895,7 @@ class CheckpointManager(
 
   def reached_preemption(self, step: int) -> bool:
     """Returns True if a preemption sync point has been reached."""
-    return utils.reached_preemption(step)
+    return multihost.reached_preemption(step)
 
   def _global_max(self, values: list[int]) -> list[int]:
     """Returns the global max of local values across all processes as a scalar.
@@ -1065,7 +1039,6 @@ class CheckpointManager(
       self,
       step: int,
       restoring_slice_id: int,
-      args: Optional[args_lib.CheckpointArgs] = None,
       directory: Optional[epath.PathLike] = None,
   ) -> Any:
     logging.info(
@@ -1077,33 +1050,11 @@ class CheckpointManager(
     step_stats = step_statistics.EmergencyRestoreStepStatistics()
     step_stats.checkpoint_manager_start_time = time.time()
     step_stats.step = step
-
     is_restoring_slice = restoring_slice_id == self._slice_id
     step_stats.is_restoring_slice = is_restoring_slice
     step_stats.in_primary_slice = self.in_primary_slice
 
     shape_dtypes, tree_defs = jax.tree.flatten(self._abstract_state)
-
-    def _get_single_slice_sharding(
-        mesh: jax.sharding.Mesh,
-        pspec: jax.sharding.PartitionSpec,
-        replica_id: int,
-        replica_axis_index: int,
-    ):
-      slice_devices = multislice.slice_devices(
-          mesh,
-          replica_id=replica_id,
-          replica_axis_index=replica_axis_index,
-      )
-      single_slice_mesh_shape = [
-          1 if i == replica_axis_index else d
-          for i, d in enumerate(mesh.devices.shape)
-      ]
-      slice_mesh = jax.sharding.Mesh(
-          slice_devices.reshape(single_slice_mesh_shape), mesh.axis_names
-      )
-      return jax.sharding.NamedSharding(slice_mesh, pspec)
-
     original_single_slice_shardings = jax.tree.map(
         lambda arr: _get_single_slice_sharding(
             mesh=arr.sharding.mesh,
@@ -1117,37 +1068,38 @@ class CheckpointManager(
         jax.tree.flatten(original_single_slice_shardings)[0]
     )
 
+    # Debug logging for sharding information.
+    if logging.vlog_is_on(1):
+      logging.vlog(
+          1,
+          'Debugging global restore_args based on abstract state. This uses the'
+          ' user-provided mesh, and is not actually used for restoration.',
+      )
+      local_checkpoint_data_debugging.print_devices_indices_debug_info(
+          checkpoint_utils.construct_restore_args(self._abstract_state)
+      )
+      logging.vlog(
+          1,
+          'Debugging single-slice restore_args based on abstract state. This'
+          ' uses the user-provided mesh, and is not actually used for'
+          ' restoration.',
+      )
+      local_checkpoint_data_debugging.print_devices_indices_debug_info(
+          checkpoint_utils.construct_restore_args(
+              self._abstract_state, original_single_slice_shardings
+          )
+      )
+
     if is_restoring_slice:
       logging.vlog(
           1, 'emergency.CheckpointManager: restoring from local checkpoint.'
       )
-      # Each device has its own specific local shard.
-      # Naive example:
-      #  Before restart: device X (with id 0) saves shard 0.
-      #  After restart:  device X (with new id 1) reads shard 0.
-      # To ensure the same device X (with different software ids) reads the
-      # same locally available shard and represent it in the correct index
-      # within the global jax.Array, we use the `restore_mesh`.
-
-      # More context on `restore_mesh`:
-      # 1. We can think of mesh as being backed by a list of devices.
-      # 2. Default mesh follows the default device id order [0, ..., n-1]. Or
-      #    the user may permute it according to their needs.
-      # 3. After restart, the user will construct the same software mesh as (2).
-      # 4. But a given hardware device may change its id because of scheduler
-      #    or runtime quirks.
-      # 5. Goal: construct the mesh with the same hardware device order as
-      #    before restart, that may not follow the current software ids.
-      # 5. Thus, we shuffle the device order within the mesh by checking how
-      #    each device's software ids changed across restarts.
-      restore_mesh = self._global_mesh
-      if _should_restore_mesh_from_metadata(self._persistent_directory):
-        logging.info(
-            'Found consistent_restore_mesh, using it for local restoration'
-        )
-        restore_mesh = _consistent_restore_mesh_from_metadata(
-            self._persistent_directory, self._global_mesh
-        )
+      restore_mesh = mesh_consistency.consistent_restore_mesh_from_metadata(
+          self._local_directory,
+          step,
+          self._global_mesh,
+          multihost.distributed_to_device_ids(),
+      )
 
       restoring_processes = multihost.unique_processes_from_devices(
           multislice.slice_devices(
@@ -1172,13 +1124,8 @@ class CheckpointManager(
           ),
           self._abstract_state,
       )
-      restore_args = jax.tree.map(
-          lambda restore_shard, arr: type_handlers.ArrayRestoreArgs(
-              sharding=restore_shard,
-              global_shape=arr.shape,  # single-slice sharding
-          ),
-          restore_single_slice_shardings,
-          self._abstract_state,
+      single_slice_restore_args = checkpoint_utils.construct_restore_args(
+          self._abstract_state, restore_single_slice_shardings
       )
       restore_directory = self._options.step_name_format.find_step(
           epath.Path(directory or self._local_directory), step
@@ -1195,10 +1142,29 @@ class CheckpointManager(
           'Restoring from %s',
           restore_directory / checkpoint_manager.DEFAULT_ITEM_NAME,
       )
+      if logging.vlog_is_on(1):
+        asyncio.run(
+            local_checkpoint_data_debugging.print_chunk_debug_info(
+                restore_directory / checkpoint_manager.DEFAULT_ITEM_NAME,
+                single_slice_restore_args,
+            )
+        )
+        local_checkpoint_data_debugging.print_devices_indices_debug_info(
+            single_slice_restore_args
+        )
+
       step_stats.checkpointer_start_time = time.time()
+      args = args_lib.PyTreeRestore(
+          item=self._abstract_state,
+          restore_args=checkpoint_utils.construct_restore_args(
+              self._abstract_state
+          ),
+      )
       single_slice_pytree = local_state_handler.restore(
           restore_directory / checkpoint_manager.DEFAULT_ITEM_NAME,
-          args=dataclasses.replace(args, restore_args=restore_args),
+          args=dataclasses.replace(
+              args, restore_args=single_slice_restore_args
+          ),
       )
       step_stats.checkpointer_duration_secs = (
           time.time() - step_stats.checkpointer_start_time
@@ -1212,7 +1178,7 @@ class CheckpointManager(
         # available shard correctly. This may cause performance issues.
 
         # Thus, we re-shard the array to follow the original mesh and layout.
-        in_tree = self._consistent_restore_mesh_to_global_mesh(
+        in_tree = mesh_consistency.consistent_restore_mesh_to_global_mesh(
             in_tree, original_single_slice_shardings_tuple
         )
     else:
@@ -1260,35 +1226,9 @@ class CheckpointManager(
 
     return jax.tree.unflatten(tree_defs, shared_states)
 
-  def _consistent_restore_mesh_to_global_mesh(
-      self,
-      original_state: PyTree,
-      desired_slice_shardings,
-  ) -> Any:
-    """Transfers from consistent restore mesh to global mesh."""
-    logging.info('Transferring from consistent restore mesh to global mesh')
-
-    start_transfer = time.time()
-    resharded_state = jax.device_put(
-        original_state, desired_slice_shardings, donate=True
-    )
-    transfer_elapsed_s = time.time() - start_transfer
-    logging.info(
-        'Finished transferring from consistent restore mesh to global mesh'
-        ' in %.2fs',
-        transfer_elapsed_s,
-    )
-    jax.monitoring.record_event_duration_secs(
-        '/orbax/emergency/checkpoint/read/transfer_global_shard_duration_secs',
-        transfer_elapsed_s,
-    )
-
-    return resharded_state
-
   def _restore_from_persistent(
       self,
       step: int,
-      args: Optional[args_lib.CheckpointArgs] = None,
       directory: Optional[epath.PathLike] = None,
   ) -> Any:
     logging.info(
@@ -1297,6 +1237,12 @@ class CheckpointManager(
         step,
         directory or self._persistent_directory,
     )
+    args = args_lib.PyTreeRestore(
+        item=self._abstract_state,
+        restore_args=checkpoint_utils.construct_restore_args(
+            self._abstract_state
+        ),
+    )
 
     # Create a temporarily read-only PersistentCheckpointManager that will
     # synchronize the restoration with global processes.
@@ -1329,12 +1275,17 @@ class CheckpointManager(
 
   def restore(
       self,
-      step: int,
+      step: Optional[int],
       args: Optional[args_lib.CheckpointArgs] = None,
       directory: Optional[epath.PathLike] = None,
   ) -> Any:
+    del args
     if step is None:
-      raise ValueError('Step must be provided for restore. Found None.')
+      step = self.latest_step()
+      if step is None:
+        raise FileNotFoundError(
+            'No steps found in persistent or local storage.'
+        )
     logging.info('Restoring at step %d.', step)
     restoring_slice_id = self._find_slice_with_complete_local_checkpoint(step)
     if restoring_slice_id > -1:
@@ -1342,13 +1293,10 @@ class CheckpointManager(
       return self._restore_from_local(
           step=step,
           restoring_slice_id=restoring_slice_id,
-          args=args,
           directory=directory,
       )
 
-    return self._restore_from_persistent(
-        step=step, args=args, directory=directory
-    )
+    return self._restore_from_persistent(step=step, directory=directory)
 
   def item_metadata(self, step: int) -> Any:
     raise NotImplementedError(
