@@ -12,8 +12,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""To test Orbax in single-host setup."""
-
 from absl.testing import absltest
 from absl.testing import parameterized
 from etils import epath
@@ -23,6 +21,7 @@ import ml_dtypes
 import numpy as np
 from orbax.checkpoint import test_utils
 from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
+from orbax.checkpoint._src.handlers import standard_checkpoint_handler_test_utils
 from orbax.checkpoint._src.serialization import type_handlers
 import tensorstore as ts
 
@@ -147,6 +146,19 @@ class SingleHostTest(parameterized.TestCase):
     np.testing.assert_array_equal(x, restored_tree['x'])
     assert isinstance(restored_tree['x'], jax.Array)
 
+  @parameterized.parameters({'x': [1, 2]}, {'x': 1})
+  def test_save_singular_array_with_standard_checkpoint_handler(self, x):
+    if isinstance(x, list):
+      x = jnp.array(x)
+    handler = standard_checkpoint_handler_test_utils.StandardCheckpointHandler()
+    with self.assertRaisesRegex(
+        ValueError, '.*Use ArrayCheckpointHandler / ArraySave.*'
+    ):
+      handler.save(
+          self.ckpt_dir,
+          args=standard_checkpoint_handler_test_utils.StandardSaveArgs(x),
+      )
+
   @parameterized.product(
       dtype=[
           jnp.bfloat16,
@@ -187,6 +199,34 @@ class SingleHostTest(parameterized.TestCase):
     np.testing.assert_array_equal(x, restored_tree['x'])
     self.assertIsInstance(restored_tree['x'], jax.Array)
 
+  def test_save_and_restore_with_replica_parallel(self):
+    if len(jax.devices()) <= 1:
+      self.skipTest('Test requires multiple devices.')
+    mesh = jax.sharding.Mesh(jax.devices(), ('x',))
+    replicated_spec = jax.sharding.PartitionSpec()
+    sharding = jax.sharding.NamedSharding(mesh, replicated_spec)
+
+    key = jax.random.PRNGKey(0)
+    state = jax.random.normal(key, (1024, 1024))
+    state = jax.device_put(state, sharding)
+
+    pytree = {'state': state}
+    array_handler = type_handlers.ArrayHandler(
+        replica_id=0,
+        use_replica_parallel=True,
+    )
+    handler = PyTreeCheckpointHandler(
+        type_handler_registry=type_handlers.create_type_handler_registry(
+            (jax.Array, array_handler),
+        ),
+    )
+    handler.save(self.ckpt_dir, pytree)
+
+    restored_tree = handler.restore(self.ckpt_dir)
+
+    np.testing.assert_array_equal(state, restored_tree['state'])
+    self.assertIsInstance(restored_tree['state'], jax.Array)
+
 
 if __name__ == '__main__':
   absltest.main()
