@@ -21,12 +21,10 @@ from __future__ import annotations
 import asyncio
 from concurrent import futures
 import contextlib
-import functools
 import inspect
-import itertools
 import time
 import typing
-from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple
+from typing import Any, List, Optional, Set
 from unittest import mock
 
 from absl import logging
@@ -39,33 +37,18 @@ from orbax.checkpoint import checkpoint_args
 from orbax.checkpoint._src.handlers import async_checkpoint_handler
 from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
 from orbax.checkpoint._src.metadata import checkpoint as checkpoint_metadata
-from orbax.checkpoint._src.multihost import counters
+from orbax.checkpoint._src.metadata import step_metadata_serialization
 from orbax.checkpoint._src.multihost import multihost
 from orbax.checkpoint._src.multihost import multislice
 from orbax.checkpoint._src.path import atomicity
 from orbax.checkpoint._src.path import step as step_lib
+from orbax.checkpoint._src.serialization import replica_slices
 from orbax.checkpoint._src.serialization import serialization
 from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
 from orbax.checkpoint._src.serialization import type_handlers
 from orbax.checkpoint._src.tree import utils as tree_utils
 
 
-class MuNu(NamedTuple):
-  mu: jax.Array | None
-  nu: np.ndarray | None
-
-
-class EmptyNamedTuple(NamedTuple):
-  pass
-
-
-class NamedTupleWithNestedAttributes(NamedTuple):
-  nested_mu_nu: MuNu | None = None
-  nested_dict: Dict[str, jax.Array] | None = None
-  nested_tuple: Tuple[str, np.ndarray] | None = None
-  nested_empty_named_tuple: EmptyNamedTuple | None = None
-
-
 def sync_global_processes(name: str):
   jax.experimental.multihost_utils.sync_global_devices(name)
 
@@ -112,9 +95,7 @@ def create_tmp_directory(
     active_processes: Optional[Set[int]] = None,
     barrier_sync_key_prefix: Optional[str] = None,
     path_permission_mode: int = step_lib.WORLD_READABLE_MODE,
-    checkpoint_metadata_store: Optional[
-        checkpoint_metadata.CheckpointMetadataStore
-    ] = None,
+    metadata_store: Optional[checkpoint_metadata.MetadataStore] = None,
 ) -> epath.Path:
   """Creates a non-deterministic tmp directory for saving for given `final_dir`.
 
@@ -130,9 +111,8 @@ def create_tmp_directory(
       0o750. Please check
       https://github.com/google/etils/blob/main/etils/epath/backend.py if your
         path is supported.
-    checkpoint_metadata_store: optional `CheckpointMetadataStore` instance. If
-      present then it is used to create `CheckpointMetadata` with current
-      timestamp.
+    metadata_store: optional `MetadataStore` instance. If present then it is
+      used to create `StepMetadata` with current timestamp.
 
   Returns:
     The tmp directory.
@@ -146,7 +126,7 @@ def create_tmp_directory(
       multihost.unique_barrier_key(
           'create_tmp_directory:pre',
           prefix=barrier_sync_key_prefix,
-          suffix=f'{final_dir.name}.{counters.tmp_directory_counter()}',
+          suffix=f'{final_dir.name}',
       ),
       timeout=multihost.DIRECTORY_CREATION_TIMEOUT,
       processes=active_processes,
@@ -167,19 +147,20 @@ def create_tmp_directory(
         )
     logging.info('Creating tmp directory %s', tmp_dir)
     tmp_dir.mkdir(parents=True, exist_ok=False, mode=path_permission_mode)
-    if checkpoint_metadata_store is not None:
-      checkpoint_metadata_store.write(
-          checkpoint_path=tmp_dir,
-          checkpoint_metadata=checkpoint_metadata.StepMetadata(
-              init_timestamp_nsecs=time.time_ns()
-          ),
+    if metadata_store is not None:
+      metadata = checkpoint_metadata.StepMetadata(
+          init_timestamp_nsecs=time.time_ns(),
+      )
+      metadata_store.write(
+          file_path=checkpoint_metadata.step_metadata_file_path(tmp_dir),
+          metadata=step_metadata_serialization.serialize(metadata),
       )
 
   multihost.sync_global_processes(
       multihost.unique_barrier_key(
           'create_tmp_directory:post',
           prefix=barrier_sync_key_prefix,
-          suffix=f'{final_dir.name}.{counters.tmp_directory_counter()}',
+          suffix=f'{final_dir.name}',
       ),
       timeout=multihost.DIRECTORY_CREATION_TIMEOUT,
       processes=active_processes,
@@ -559,32 +540,12 @@ def _get_test_wrapper_function(test_func):
     def _get_unique_barrier_key(key: str) -> str:
       return f'{key}.{self.id()}'
 
-    def _patched_save_counter(test_counter) -> str:
-      return f'{self.id()}_{next(test_counter)}'
-
-    test_counters = {}
-    patched_counters = {}
-    for name, _ in inspect.getmembers(counters, predicate=inspect.isfunction):
-      test_counters[name] = itertools.count()
-      patched_counters[name] = functools.partial(
-          _patched_save_counter, test_counters[name]
-      )
-
     with mock.patch.object(
         multihost,
         '_unique_barrier_key',
         new=_get_unique_barrier_key,
     ):
-      with contextlib.ExitStack() as stack:
-        for name, patched_counter in patched_counters.items():
-          stack.enter_context(
-              mock.patch.object(
-                  counters,
-                  name,
-                  new=patched_counter,
-              )
-          )
-        return test_func(self, *args, **kwargs)
+      return test_func(self, *args, **kwargs)
 
   return test_wrapper
 
@@ -619,22 +580,20 @@ def barrier_compatible_test(cls):
 def ensure_atomic_save(
     temp_ckpt_dir: epath.Path,
     final_ckpt_dir: epath.Path,
-    checkpoint_metadata_store: Optional[
-        checkpoint_metadata.CheckpointMetadataStore
-    ] = None,
+    metadata_store: Optional[checkpoint_metadata.MetadataStore] = None,
 ):
   """Wrapper around TemporaryPath.finalize for testing."""
   if temp_ckpt_dir == final_ckpt_dir:
     atomicity.CommitFileTemporaryPath(
         temp_ckpt_dir,
         final_ckpt_dir,
-        checkpoint_metadata_store=checkpoint_metadata_store,
+        checkpoint_metadata_store=metadata_store,
     ).finalize()
   else:
     atomicity.AtomicRenameTemporaryPath(
         temp_ckpt_dir,
         final_ckpt_dir,
-        checkpoint_metadata_store=checkpoint_metadata_store,
+        checkpoint_metadata_store=metadata_store,
     ).finalize()
 
 
@@ -751,3 +710,13 @@ def assert_every_n_is_x_apart(testclass, values, n, x):
   assert len(values) % n == 0
   for i in range(n, len(values), n):
     testclass.assertGreaterEqual(values[i], values[i - 1] + x)
+
+
+def get_expected_chunk_shape(arr: jax.Array) -> tuple[int, ...]:
+  """Expected chunk shape for an array, accounting for replica-parallel."""
+  _, local_shape = (
+      replica_slices.calculate_replica_parallel_axis_and_local_shape(arr)
+  )
+  if local_shape is None:
+    local_shape = arr.sharding.shard_shape(arr.shape)
+  return local_shape
