@@ -18,7 +18,6 @@ from absl.testing import absltest
 from absl.testing import parameterized
 from etils import epath
 from orbax.checkpoint import test_utils
-from orbax.checkpoint._src.multihost import counters
 from orbax.checkpoint._src.multihost import multihost
 
 
@@ -40,16 +39,6 @@ class BarrierCompatibleTest(parameterized.TestCase):
         multihost._unique_barrier_key('footest_path_permission_mode'),
     )
 
-  def test_unique_counter_primary(self):
-    expected_key = 'BarrierCompatibleTest.test_unique_counter_primary'
-    self.assertIn(f'{expected_key}_0', counters.async_save_counter())
-    self.assertIn(f'{expected_key}_1', counters.async_save_counter())
-
-  def test_unique_counter_secondary(self):
-    expected_key = 'BarrierCompatibleTest.test_unique_counter_secondary'
-    self.assertIn(f'{expected_key}_0', counters.async_save_counter())
-    self.assertIn(f'{expected_key}_1', counters.async_save_counter())
-
 
 if __name__ == '__main__':
   absltest.main()
